/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Abacus Template Editor.
 *
 * The Initial Developer of the Original Code is 
 * Alexander J. Vincent <ajvincent@juno.com>.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

function abacusTemplateEditor() {
  this.templatesTable = new abacusTemplatesTable();
  this.testMathElement = null;
  this.testDocTxMgr = Components.classes["@mozilla.org/transactionmanager;1"].createInstance(nsITransactionManager);
}
abacusTemplateEditor.prototype = {
  fields: ["operationName", "indexCode", "xml:lang", "markupLanguage", "alternateIndex"],
  grid: null,
  dialogs: {},
  userErrorBox: null,
  menubarDeck: null,

  watchUserErrors: function(methodName) {
    assert(this.userErrorBox, "There must be a user error box in place!", true);
    var args = [];
    for (var i = 1; i < arguments.length; i++) {
      args[i - 1] = arguments[i];
    }
    this.userErrorBox.value = "";
    try {
      if (ecmaDebug) {
        args.unshift(this);
        args.unshift(this[methodName]);
        applyContract.apply(window, args); // we might as well enforce contracts when available
      } else {
        this[methodName].apply(this, args);
      }
    }
    catch(e) {
      if (e.name == "UserError") {
        this.userErrorBox.value = e.message;
      } else {
        throw e;
      }
    }
  },

  init: function() {
    var i;
    var prop;
    var menu;

    // initialize user error box
    var userErrorBoxes = document.getElementsByTagName("usererrorbox");
    assert(userErrorBoxes.length == 1, "There must be exactly one usererrorbox element!", true);
    this.userErrorBox = userErrorBoxes[0];

    // initialize grid
    this.grid = document.getElementById("editorGrid");
    this.grid.searchTemplateRow = document.getElementById("searchTemplateRow");
    this.grid.newTemplateRow = document.getElementById("newTemplateRow");
    this.grid.columns = [];
    this.grid.menus = [];

    // initialize grid columns and menus
    var tableFieldValues = [];
    for (i = 0; i < this.fields.length; i++) {
      prop = this.fields[i];
      menu = document.getElementById(prop);

      // initialize column index for grid
      var j = -1;
      var currentNode = menu.parentNode;
      assert(currentNode.localName == "column", "currentNode is not a column in the grid", true);
      while (currentNode) {
        j++;
        currentNode = currentNode.previousSibling;
      }
      this.grid.columns[prop] = j;
      this.grid.menus[j] = menu;
      this.grid.menus[prop] = menu;
    }

    this.templatesTable.addFile("chrome://abacus/content/templates/base.xml", abacusTemplateFile.BASE_FILE);
    this.templatesTable.addFile("chrome://abacus/content/templates/math.xml", abacusTemplateFile.MATH_FILE);

    for (i = 0; i < this.fields.length; i++) {
      prop = this.fields[i];
      menu = this.grid.menus[prop];
      var menuPopup = menu.firstChild;
      if ((!menuPopup)||(menuPopup.childNodes.length != 0)) {
        continue; // we don't want to populate a control that's already filled!
      }

      // initialize menu popup for column
      tableFieldValues[prop] = this.templatesTable.dumpColumnValues(prop);
      for (j = 0; j < tableFieldValues[prop].length; j++) {
        var field = tableFieldValues[prop][j];
        var menuitem = document.createElement("menuitem");
        menuitem.setAttribute("label", field);
        menuPopup.appendChild(menuitem);
      }
    }
    uiCommands.init(this.templatesTable.userInterfaces);
    this.testMathElement = document.getElementById("testMathElement");
    var errMsg = "I don't have a testMathElement to run through the MathML editor.";
    assert(this.testMathElement, errMsg, true);
  },

  sortSearchResults: function(a, b) {
    var fields = editor.fields;
    for (var i = 0; i < fields.length; i++) {
      if (a[fields[i]] < b[fields[i]]) {
        return -1;
      }
      if (a[fields[i]] > b[fields[i]]) {
        return 1;
      }
    }
    return 0;
  },

  sanityCheckInput: function() {
    warn("We don't yet check operationName to ensure it follows the letters, colon, letters format.");
    // sanity-check user input (not a precondition per the applyContract() call, as that requires ecmaDebug == true)
    var templateFileName = "math";
    var operationName = this.grid.menus["operationName"].value.toString();
    if (operationName.indexOf(":") != -1) {
      templateFileName = operationName.substr(0, operationName.indexOf(":"));
      var errMsg = "The template editor currently supports only base:base and math templates.";
      errMsg += "  (This may change in a future version.)";
      assertAboutInput(templateFileName == "base", errMsg, true);

/*** If I enable creating templates for other languages, I'll probably use this section.
      assertAboutInput(templateFileName != "", "You must have a prefix with a colon in the operation name.", true);
      assertAboutInput(templateFileName != "math", "The 'math' prefix is not needed for math expressions.", true);
      assertAboutInput(templateFileName != "summary", "The 'summary' prefix is reserved!", true);
      assertAboutInput(templateFileName != "baseTemplate", "The 'baseTemplate' prefix is reserved!", true);
*/
    }

    if (templateFileName == "base") {
      errMsg = "The base prefix is reserved for base:base alone!";
      errMsg += "  (This may change in a future version.)"; 
      assertAboutInput(operationName == "base:base", errMsg, true);
    }

    var menu = this.grid.menus["alternateIndex"];
    assertAboutInput(!menu.invalid, "You cannot create a new template with an invalid alternateIndex value.", true);

    return templateFileName + ".xml";
  },

  search: function() {
    var i;
    var j;
    var searchCriteria;
    var prop;
    var menu;
    var row;
    var searchResults;
    var passesSearch;
    var criterium;
    var resultTemplate;
    var secondaryTemplate;
    var currentTemplate;
    var resultRow;
    var columnNumber;
    var resultCell;
    var templateFileName;

    
    this.sanityCheckInput();
    // after this point, there should appear no assertAboutInput() calls.  We assume past this point user input is good.

    // set search criteria
    searchCriteria = [];
    for (prop in this.grid.columns) {
      menu = this.grid.menus[prop];
      if (prop == "indexCode") {
        if (!(/\S/.test(this.grid.menus["operationName"].value))) {
          continue;
        }
      }

      if (prop == "alternateIndex") {
        if (!(/\S/.test(this.grid.menus["xml:lang"].value))) {
          continue;
        }
        if (!(/\S/.test(this.grid.menus["markupLanguage"].value))) {
          continue;
        }

      }
      if (/\S/.test(menu.value)) { // this still works with <mEdit:integerControl/>
        searchCriteria.push({ name: prop, value: menu.value, toString: function() {return this.name + ' = ' + this.value}});
      }
    }

    // get primary search results
    searchResults = [];
    for (i = 0; i < this.templatesTable.rows.length; i++) {
      row = this.templatesTable.rows[i];
      if (this.templatesTable.meetsCriteria(searchCriteria, row)) {
        searchResults.push(row);
      }
    }

    // get secondary search results (content templates for results which generated presentation templates)
    for (i = 0; i < searchResults.length; i++) {
      resultTemplate = searchResults[i];
      if (resultTemplate.isContentTemplate) {
        continue;
      }
      secondaryTemplate = null;
      for (j = 0; j < this.templatesTable.rows.length; j++) {
        currentTemplate = this.templatesTable.rows[j];
        if (!currentTemplate.isContentTemplate) {
          continue;
        }
        if (currentTemplate.operationName != resultTemplate.operationName) {
          continue;
        }
        if (currentTemplate.indexCode != resultTemplate.indexCode) {
          continue;
        }
        secondaryTemplate = currentTemplate;
        break;
      }
      assert(secondaryTemplate, "Content template for corresponding presentation template not found!", true);
      searchResults.push(secondaryTemplate);
    }

    // sort search results and remove duplicates
    searchResults.sort(this.sortSearchResults);
    for (i = 0; i < searchResults.length - 1;) {
      if (searchResults[i] == searchResults[i + 1]) {
        searchResults.splice(i + 1, 1);
      } else {
        i++;
      }
    }

    // clear previous search results
    while (this.grid.searchTemplateRow.previousSibling != this.grid.newTemplateRow) {
      this.grid.searchTemplateRow.parentNode.removeChild(this.grid.searchTemplateRow.previousSibling);
    }

    // display results
    for (i = 0; i < searchResults.length; i++) {
      resultTemplate = searchResults[i];
      resultRow = this.grid.searchTemplateRow.cloneNode(true);
      resultRow.removeAttribute("id"); // to prevent problems later
      j = 0;
      for (prop in this.grid.columns) {
        columnNumber = this.grid.columns[prop];
        resultCell = resultRow.childNodes[columnNumber];
        resultCell.appendChild(document.createTextNode(resultTemplate[prop]));
        j++;
      }
      assert(j == this.fields.length, "editor.grid.columns.length != editor.fields.length!", true);
      resultRow.selectedTemplate = resultTemplate;
      this.grid.searchTemplateRow.parentNode.insertBefore(resultRow, this.grid.searchTemplateRow);
      resultRow.hidden = false;
    }

    // display new template row;
    this.grid.newTemplateRow.hidden = false;
  },

  testField: function(aFieldName, aMustMatch, aErrMsg) {
    var rv = /\S/.test(this.grid.menus[aFieldName].value);
    if (aErrMsg) {
      assertAboutInput(rv == aMustMatch, aErrMsg, true);
    }
    return rv;
  },

  addNewTemplate: function() {
    var mustBeTrue;
    var response = true;
    var requiredTemplatesCode;
    var searchCriteria;
    var row;
    var result;
    var baseNode;
    var contentNode;
    var presentationNode;
    var newContentTemplate;
    var newPresentationTemplate;
    /* requirements:
       Content Set:    Unconditionally required
         operationName:  no default
         indexCode    :  default = "A"

       Presentation Set: Either included totally or omitted (though UI offers alternateIndex anyway)
         xml:lang:  no default
         markupLanguage: no default
         alternateIndex: default = 0
    */

    // check content fields;
    this.testField("operationName", true, "operationName field must not be empty.");
    var indexValue = this.grid.menus["indexCode"].value;

    // check presentation fields
    if (this.testField("xml:lang", true, null)) {
      this.testField("markupLanguage", true, "xml:lang value requires a markupLanguage value.");
      requiredTemplatesCode = 1;
    } else {
      this.testField("markupLanguage", false, "markupLanguage value requires a xml:lang value.");
      requiredTemplatesCode = 0;
    }

    var templateFileName = this.sanityCheckInput();
    // after this point, there should appear no assertAboutInput() calls.  We assume past this point user input is good.

    var aTemplateFile = this.templatesTable.getTemplateFileByName(templateFileName);

    // find out if we have a content markup template for a requested presentation markup template
    if (requiredTemplatesCode == 1) {
      searchCriteria = [
        {name: "dialogTitle", value: this.grid.menus["operationName"].value + "/" + this.grid.menus["indexCode"].value}
      ];
      result = this.templatesTable.getFirstSearchResult(searchCriteria);
      if (result) {
        requiredTemplatesCode++;
      }
    }

    /* requiredTemplatesCode:
       value == 0:  this is a new content template, and requires a new presentation template.
       value == 1:  this is a new presentation template for which a content template does not exist.
       value == 2:  this is a new presentation template for which a content template already exists.
    */

    // start initializing basic template information
    baseNode = document.createElementNS(MEDIT_NS + "#templates", "template");
    baseNode.setAttribute("operationName", this.grid.menus["operationName"].value);
    baseNode.setAttribute("indexCode", this.grid.menus["indexCode"].value);

    if (this.grid.menus["operationName"].value == "base:base") {
      // base:base elements require a prefix and a postfix
      var prefixNode = document.createElementNS(MEDIT_NS + "#templates", "basePrefix");
      var postfixNode = document.createElementNS(MEDIT_NS + "#templates", "basePostfix");
      baseNode.appendChild(prefixNode);
      baseNode.appendChild(postfixNode);
    }

    // create a new content template
    if ((requiredTemplatesCode == 0)||(requiredTemplatesCode == 1)) {
      contentNode = baseNode.cloneNode(true);
      contentNode.targetFileName = templateFileName;
      newContentTemplate = applyContractConstructor(abacusTemplate, contentNode, aTemplateFile, true);
      newContentTemplate.templateEditCode = 4; // per the abacusTemplate() function
    }
    
    // initialize presentation template information
    switch (requiredTemplatesCode) {
      case 0:
        baseNode.setAttributeNS(XML_NS, "xml:lang", "en-US");
        baseNode.setAttribute("markupLanguage", "MathML");
        baseNode.setAttribute("alternateIndex", "0");
        break;
      case 1:
      case 2:
        baseNode.setAttributeNS(XML_NS, "xml:lang", this.grid.menus["xml:lang"].value);
        baseNode.setAttribute("markupLanguage", this.grid.menus["markupLanguage"].value);
        baseNode.setAttribute("alternateIndex", this.grid.menus["alternateIndex"].value);
    }
    presentationNode = baseNode;
    presentationNode.targetFileName = templateFileName;
    newPresentationTemplate = applyContractConstructor(abacusTemplate, presentationNode, aTemplateFile, false);
    newPresentationTemplate.templateEditCode = 4; // per the abacusTemplate() function

    // open dialogs
    switch (requiredTemplatesCode) {
      case 0:
      case 1:
        var newContentDialog = this.openTemplateDialog(newContentTemplate); // get dialog (whose template may already exist)
        var newPresentationDialog = this.openTemplateDialog(newPresentationTemplate); // ditto
        if (typeof newContentDialog.aTemplate == "object") {
          newContentTemplate = newContentDialog.aTemplate // scrap the object we don't care to use;
        }
        if (typeof newPresentationDialog.aTemplate == "object") {
          newPresentationTemplate = newPresentationDialog.aTemplate // scrap the object we don't care to use;
        }
        newContentTemplate.addMatchingTemplate(newPresentationTemplate); // register matched template
        newPresentationTemplate.matchingTemplates = newContentTemplate.matchingTemplates; // copy matches
        /*
          You may wonder why I do it this roundabout way.  If I open:
          foo:fighters/A/en-US/MathML/0

          as a new template, and foo:foo/A doesn't exist, it will create a foo:fighters/A for me.
          Then, if I open:

          foo:fighters/A/en-US/XHTML/0
          we want to bind all the matching templates together as one set.  
          Canceling one of them cancels all of them.          
        */

        break;

      case 2:
        this.openTemplateDialog(newPresentationTemplate);
    }
  },

  setBaseTemplate: function(aTemplate) {
    var searchCriteria;
    var result;
    if (aTemplate.isContentTemplate) {
      searchCriteria = [{name: "dialogTitle", value: "base:base/A"}];
      result = this.templatesTable.getFirstSearchResult(searchCriteria);
      assert(result, "base:base/A template was not found!  Did you load base.xml as a templates file?", true);
    } else {
      searchCriteria = [
        {name: "operationName",  value: "base:base"},
        {name: "indexCode",      value: "A"},
        {name: "xml:lang",       value: this.grid.menus["xml:lang"].value},
        {name: "markupLanguage", value: this.grid.menus["markupLanguage"].value},
        {name: "alternateIndex", value: "0"}
      ];
      result = this.templatesTable.getFirstSearchResult(searchCriteria);  // this is the ideal base template

      if (!result) {
        // try the default markup language
        searchCriteria[3] = {name: "markupLanguage", value: "MathML"};
        result = this.templatesTable.getFirstSearchResult(searchCriteria);
      }

      if (!result) {
        // try the default xml:lang
        searchCriteria[2] = {name: "xml:lang", value: "en-US"};
        searchCriteria[3] = {name: "markupLanguage", value: this.grid.menus["markupLanguage"].value};
        result = this.templatesTable.getFirstSearchResult(searchCriteria);
      }

      if (!result) {
        // try the default xml:lang and markup language.  If this fails, something is really wrong!
        searchCriteria[3] = {name: "markupLanguage", value: "MathML"};
        result = this.templatesTable.getFirstSearchResult(searchCriteria);
      }
      var presentationNotFound = "No presentation base template (even base:base/A/en-US/MathML/0) was found!";
      presentationNotFound += "  Did you load base.xml?"
      assert(result, presentationNotFound, true);
    }
    aTemplate.baseTemplate = result;
  },

  dialogOffset: -20,

  openTemplateDialog: function(aTemplate) {

    if (!aTemplate.baseTemplate) {
      this.setBaseTemplate(aTemplate);
    }
    var aURL;
    assert(aTemplate.templateEditCode != 5,
           "How did a canceled template (" + aTemplate.dialogTitle + ") request opening?", true);
    if ((typeof this.dialogs[aTemplate.dialogTitle] == "undefined")||(this.dialogs[aTemplate.dialogTitle].closed)) {
      this.dialogOffset = (this.dialogOffset + 20) % 100
      var params = "chrome,width=700,height=500,resizable=1,top=" + this.dialogOffset + ",left=" + this.dialogOffset;
      if (aTemplate.isBase) {
        aURL = "chrome://abacus/content/templates/baseTemplateDialog.xul";
      } else {
        aURL = "chrome://abacus/content/templates/templateDialog.xul";
      }
      this.dialogs[aTemplate.dialogTitle] = window.openDialog(aURL, aTemplate.dialogTitle, params, aTemplate);
    }
    this.dialogs[aTemplate.dialogTitle].focus();
    return this.dialogs[aTemplate.dialogTitle];
  },

  templateCanceled: function(aTemplate) {
    if (aTemplate.templateEditCode == 4) {
      // we have a new template that doesn't exist in a file yet
      var needsUserError = false;
      for (var i = 0; i < aTemplate.matchingTemplates.length; i++) {
        var match = aTemplate.matchingTemplates[i];
        if (match == aTemplate) {
          continue;
        }
        // uh oh! this is part of a matched set... bail out of this one too
        needsUserError = true;
        var matchCleared = (match.templateEditCode != 4);
        if (!matchCleared) {
          match.templateEditCode = 5; // pre-emptively set abort code to prevent runaway recursion
          this.dialogs[match.dialogTitle].doCancel(); // shut down
        }
      }
      aTemplate.templateEditCode = 5; // note abort code
      if (needsUserError) {
        window.focus(); // so the user can see the error message
        throw new UserError("The dialog you canceled is part of a content/presentation set that has not been added.");
        // tell the user why the other dialogs disappeared suddenly
      }
    }
  // the canceled template is never added to this.templatesTable.rows, not even if it's a canceled new template.
  },

  templateUpdated: function(aTemplate) {
    if (aTemplate.templateEditCode == 3) {
      // we have a new template that doesn't exist in a file yet
      for (var i = 0; i < aTemplate.matchingTemplates.length; i++) {
        var match = aTemplate.matchingTemplates[i];
        // it's part of a matched template
        if (match.templateEditCode == 4) {
          // apparently the other template isn't finished editing yet... we'll pass for now on adding the template.
          return;
        }
        assert(match.templateEditCode == 3, 
               "Unexpected template edit code (" + match.templateEditCode + ") for template " + match.dialogTitle,
               true);
      }
      // hm, we got here... that means all the templates were finished!
      for (i = 0; i < aTemplate.matchingTemplates.length; i++) {
        match = aTemplate.matchingTemplates[i];
        this.addTemplateToTable(match);
      }
    }
  },

  addTemplateToTable: function(aTemplate) {
    assert(aTemplate.templateEditCode == 3, aTemplate.dialogTitle + " template is not ready for adding to table!", true);
    assert(aTemplate.templateFile, aTemplate.dialogTitle + " template did not have a templateFile set!", true);
    this.templatesTable.rows.push(aTemplate);
    aTemplate.templateEditCode = 1;
    aTemplate.isNewTemplate = true; // used later in aTemplate.saveContent()
  },

  saveChanges: function() {
    var i;
    var j;
    var aTemplate;
    var aTemplateFile;
    var errMsg;
    var parsedTemplateDoc;
    var aUIGroupsNode;

    errMsg = "All template dialogs must be closed before saving."
    for (i = 0; i < this.templatesTable.rows.length; i++) {
      aTemplate = this.templatesTable.rows[i];
      if (typeof this.dialogs[aTemplate.dialogTitle] == "undefined") {
        continue; // we never opened a dialog for this one in the first place
      }
      assertAboutInput(this.dialogs[aTemplate.dialogTitle].closed, errMsg, true);
    }

    // make sure each uiGroups element fixes its source code
    // this comes before the template updates because the user interfaces code is sensitive to aTemplateFile.needsSave
    uiCommands.saveContent();

    // make sure each template fixes its source code
    for (i = 0; i < this.templatesTable.rows.length; i++) {
      aTemplate = this.templatesTable.rows[i];
      switch (aTemplate.templateEditCode) {
        case 0:
          break;

        case 1:
          aTemplate.saveContent();
          break;

        default:
          errMsg = "Invalid template edit code (" + aTemplate.templateEditCode + ") at saveChanges function!\n";
          errMsg += "Warning!  Several template edit codes have been set to 0 already, but the files have not been saved!"
          assert(false, errMsg, true);
      }
    }

    // actually save the files
    for (i = 0; i < this.templatesTable.files.length; i++) {
      aTemplateFile = this.templatesTable.files[i];
      if (aTemplateFile.needsSave) {
        applyContract(aTemplateFile.save, aTemplateFile);
      }
    }

    warn("We really need some other method of notifying the user that the save was successful.");
    assertAboutInput(false, "All files saved successfully!", true)
  },

  openMathMLEditor: function() {
    var newMathNode = this.testMathElement.cloneNode(true);
    var txn = new abacusTransaction(this.testDocTxMgr, newMathNode, this.testMathElement, "urn:abacus:templateEditor:test");
    txn.reallyDoTransaction = txn.doTransaction;
    txn.doTransaction = function() {
      editor.testMathElement = this.mathNode;
      this.reallyDoTransaction();
    }
    txn.reallyUndoTransaction = txn.undoTransaction;
    txn.undoTransaction = function() {
      editor.testMathElement = this.originalContent;
      this.reallyUndoTransaction();
    }
    txn.openDialog({
      templatesTable: this.templatesTable,
//    exportIndexNodes: true,
/*
   There is currently a bug when we try to retrieve an <mEdit:fitb/> element's attributes property.

   Error: uncaught exception: 
   [Exception... "Illegal operation on WrappedNative prototype object"  
   nsresult: "0x8057000c (NS_ERROR_XPC_BAD_OP_ON_WN_PROTO)"
   location: "JS frame :: javascript:alert(document.getElementsByTagNameNS(MEDIT_NS, "fitb")[0].attributes) :: 
   <TOP_LEVEL> :: line 1"  data: no]

   Other things attribute-related also break.
 */
      toString: function() {
        return "[object abacusRuntimeConfiguration]";
      }
    });
  },

  toString: function() {
    return "[object abacusTemplateEditor]";
  }
};

/**** baseline editor functions ****/

var errorMessage;
const fUtils = new FileUtils();
const parser = new DOMParser();
parser.checkRootElement = function(aDocument, aRootElementName) {
  return (aDocument.documentElement.nodeName == aRootElementName);
}

// turn on assertions and design-by-contract validation
ecmaDebug = true;
// turn on JSLIB Errors
JS_LIB_ERROR = 1;
JS_LIB_ERROR_ALERT = ON;

// initialize our template editor
var editor = new abacusTemplateEditor();
window.addEventListener("load", function(evt) {
  try {
    editor.init();
  }
  catch(e) {
    alert(e);
  }
}, true);

