/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is ECMAScript debugging functions.
 *
 * The Initial Developer of the Original Code is
 * Alexander J. Vincent <ajvincent@juno.com>.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var nsIScriptError = Components.interfaces.nsIScriptError;
var nsIConsoleService = Components.interfaces.nsIConsoleService;
var ecmaDebug;

function dispatchMessage(aError, aMessageFlag) {
  var aMessage = Components.classes["@mozilla.org/scripterror;1"].createInstance(nsIScriptError);
  aMessage.init(aError.message, aError.fileName, "", aError.lineNumber, 0, aMessageFlag, "");
  var consoleService = Components.classes["@mozilla.org/consoleservice;1"].getService(nsIConsoleService);
  consoleService.logMessage(aMessage);
}

function warn(aIncomingError) {
  var aError;
  if (aIncomingError instanceof Error) {
    aError = aIncomingError;
  } else {
    try {
      throw new Error(aIncomingError.toString()); // we want the error initialized, including stack
    }
    catch(e) {
      aError = e;
      aError.shiftStack(2); // get to the actual warning
    }
  }
  dispatchMessage(aError, nsIScriptError.warningFlag);
}

function assert(mustBeTrue, errMsg) {
  if (!ecmaDebug) {
    return mustBeTrue;
  }
  var throwException = arguments.length > 2 ? arguments[2] : false;
  try {
    if (!mustBeTrue) {
     var aError = new Error(errMsg);
     throw aError;
    }
  }
  catch(e) {
    e.shiftStack(2) // this takes us to the actual stack error.
    e.name = "AssertionError";
    if (throwException) {
      e.message = errMsg; // reset
      throw e;
    } else {
      warn(e);
    }
  }
  return mustBeTrue;
}

Error.prototype.shiftStack = function(aShiftLines) {
  var aStackArray = this.stack.split("\n").splice(aShiftLines);
  var lastColon = aStackArray[0].lastIndexOf(":");
  this.fileName = aStackArray[0].substring(aStackArray[0].indexOf("@") + 1, lastColon);
  this.lineNumber = aStackArray[0].substr(lastColon + 1);
  this.stack = aStackArray.join("\n");
}

function applyContract(aFunction, thisObj) {
  assert(typeof aFunction == "function", "applyContract must receive a function!", true);
  var args = [];
  for (var x = 2; x < arguments.length; x++) {
    args[x - 2] = arguments[x];
  }
  if ((ecmaDebug)&&(typeof aFunction.precondition == "function")) {
    errMsg = "Preconditions must accept the same number of arguments as the main function.";
    assert(aFunction.length == aFunction.precondition.length, errMsg, true);
    aFunction.precondition.apply(this, args);
  }
  var rv = aFunction.apply(thisObj, args);
  if ((ecmaDebug)&&(typeof aFunction.postcondition == "function")) {
    errMsg = "Postconditions must accept one more argument (the return value) than the main function.";
    assert(aFunction.length + 1 == aFunction.postcondition.length, errMsg, true);
    args.push(rv);
    aFunction.postcondition.apply(this, args);
  }
  return rv;
}

function constructNew(aFunction) {
  assert(typeof aFunction == "function", "constructNew must receive a function!", true);
  var args = [];
  var argString = "";
  for (var x = 1; x < arguments.length; x++) {
    args[x - 1] = arguments[x];
    argString += ",args[" + (x - 1) + "]";
  }
  argString = argString.substr(1);
  eval("var rv = new aFunction(" + argString + ")");
  return rv;
}

function applyContractConstructor(aFunction) {
  assert(typeof aFunction == "function", "applyContractConstructor must receive a function!", true);
  var errMsg;
  var args = [];
  for (var x = 1; x < arguments.length; x++) {
    args[x - 1] = arguments[x];
  }
  if ((ecmaDebug)&&(typeof aFunction.prototype.precondition == "function")) {
    errMsg = "Preconditions must accept the same number of arguments as the main function.";
    assert(aFunction.length == aFunction.prototype.precondition.length, errMsg, true);
    aFunction.prototype.precondition.apply(this, args);
  }
  args.unshift(aFunction);
  var rv = constructNew.apply(this, args);
  assert(rv instanceof aFunction, "A new object of the intended class was not created!", true);
  args.shift(aFunction);
  if ((ecmaDebug)&&(typeof rv.postcondition == "function")) {
    errMsg = "Postconditions must accept one more argument (the return value) than the main function.";
    if (aFunction.length + 1 != rv.postcondition.length) {
      alert(aFunction);
    }
    assert(aFunction.length + 1 == rv.postcondition.length, errMsg, true);
    args.push(rv);
    rv.postcondition.apply(this, args);
  }
  return rv;
}

function setContract(aFunction, aPrecondition, aPostcondition) {
  if (!ecmaDebug) {
    return;
  }
  var errMsg = "setContract must receive a function as its first argument!";
  assert(typeof aFunction == "function", errMsg, true);
  errMsg = "The main function already has a contract!";
  assert((typeof aFunction.precondition == "undefined")&&(typeof aFunction.postcondition == "undefined"), errMsg, true);
  if (aPrecondition !== null) {
    errMsg = "The precondition argument must be either null or a function."
    assert(typeof aPrecondition == "function", errMsg, true);
    errMsg = "Preconditions must accept the same number of arguments as the main function.";
    assert(aFunction.length == aPrecondition.length, errMsg, true);
  }
  if (aPostcondition !== null) {
    errMsg = "The postcondition argument must be either null or a function."
    assert(typeof aPostcondition == "function", errMsg, true);
    errMsg = "Postconditions must accept one more argument (the return value) than the main function.";
    assert(aFunction.length + 1 == aPostcondition.length, errMsg, true);
  }
  aFunction.precondition  = aPrecondition;
  aFunction.postcondition = aPostcondition;
}
