/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Abacus mEdit Index Synchronizer.
 *
 * The Initial Developer of the Original Code is 
 * Alexander J. Vincent <ajvincent@juno.com>.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

function mEditIndexSynchronizer() {
  this.disabledArray = [];
  this.indexArray = [];
  this.rootNodes = [];
}
mEditIndexSynchronizer.prototype = {
  byAttributeNSFilter: {
    value: "-1",
    namespaceURI: null,
    localName: null,

    acceptNode: function(aNode) {
      if (aNode.getAttributeNS(this.namespaceURI, this.localName) === this.value) {
        return NodeFilter.FILTER_ACCEPT;
      }
      return NodeFilter.FILTER_SKIP;
    },

    init: function(aNamespaceURI, aLocalName, aValue) {
      this.namespaceURI = aNamespaceURI;
      this.localName    = aLocalName;
      this.value        = aValue;
    },

    toString: function() {
      return "[object NodeFilter]";
    }
  },

  addRootNode: function(aNode) {
    // check to see if aNode is already a member
    for (var i = 0; i < this.rootNodes.length; i++) {
      if (this.rootNodes[i] == aNode) {
        return;
      }
    }
    this.rootNodes.push(aNode);
  },

  getIndexValue: function(aIndex) {
    var errMsg = "mEditIndexSynchronizer.getIndexValue(" + aIndex + ") failed!";
    assert(typeof this.indexArray[aIndex] != "undefined", errMsg, true);
    return this.indexArray[aIndex];
  },


  setDisabledValue: function(aIndex, aValue) {
    this.disabledArray[aIndex] = aValue;
  },

  updateValues: function(aIndex) {
    var aValue = this.getIndexValue(aIndex);
    for (var i = 0; i < this.rootNodes.length; i++) {
      var rootNode = this.rootNodes[i];
      var walker = document.createTreeWalker(rootNode, NodeFilter.SHOW_ELEMENT, this.byAttributeNSFilter, true);
      this.byAttributeNSFilter.init(MEDIT_NS, "index", aIndex.toString());
      if (this.byAttributeNSFilter.acceptNode(rootNode) == NodeFilter.FILTER_ACCEPT) {
        rootNode.value = aValue;
      }
      while (walker.nextNode()) {
        walker.currentNode.value = aValue;
        if ((walker.currentNode.namespaceURI == MEDIT_NS)&&(walker.currentNode.localName == "fitb")) {
          if (typeof walker.currentNode.blankType == "undefined") {
            continue;
            // It's entirely possible that no binding applies to this element (display:none;)
          }
          walker.currentNode.updateIsDisabled();
        }
      }
    }
  },

  setIndexValue: function(aIndex, aValue) {
    this.indexArray[aIndex] = aValue;
    this.updateValues(aIndex);
  },

  toString: function() {
    return "[object mEditIndexSynchronizer]";
  }
};
